
/* ************************************************
 * Project	: AVR bootloader download
 * Platform	: x86_32
 * 		: gcc-4.1.2
 * 		: binutils-2.17
 * 		: glibc-2.7-2
 * 		:
 * OS		: Fedora8 (Linux 2.6.23.1-42.fc8)
 * Editor	: VIM (utf-8)
 * Author	: mowin
 * Date		: 2008-06-17
 * ***********************************************/

#include "serapp.h"

#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DEBUG 0

#define TRUE 0
#define FALSE -1

/* 从机连接码 */
#define CONNECT_CODE 'd'

/* XMODEM控制字 */
#define XMODEM_NUL 0x00
#define XMODEM_SOH 0x01
#define XMODEM_STX 0x02
#define XMODEM_EOT 0x04
#define XMODEM_ACK 0x06
#define XMODEM_NAK 0x15
#define XMODEM_CAN 0x18
#define XMODEM_EOF 0x1a
#define XMODEM_RWC 'C'

/* ************************************************
 * Function	: crc16
 * 		: 计算CRC校验(1021)
 * argument	: buf 	输入数据数组
 * 		: n	数据长度
 * return	: void
 * global var	: ch (crc16结果高字节)
 * 		: cl (crc16结果低字节)
 * ***********************************************/
unsigned char ch, cl;
void crc16(unsigned char *buf, unsigned char n)
{
    unsigned char i, j;
    unsigned int crc, t;

    crc = 0;
    for(j = n; j > 0; j--){
	crc = (crc ^ (((unsigned int) *buf) << 8));
	for(i = 8; i > 0; i--){
	    t = crc << 1;
	    if(crc & 0x8000)
		t = t ^ 0x1021;
	    crc = t;
	}
	buf++;
    }
    ch = crc / 256;
    cl = crc % 256;
}
/* ************************************************
 * Function 	: str2hex 
 * 		: 用于从命令行参数获取波特率数据
 * argument	: str 命令行波特率字符串
 * return 	: 波特率数据
 * 		: 错误返回FALSE
 * 		: 输入数据不在指定范围返回0
 * ***********************************************/
unsigned long str2hex(char *str)
{
    char len;
    char *buf;
    char i;
    unsigned long ret;

    len = strlen(str);

    buf = malloc(len);
    if(buf == NULL)
	return FALSE;

    for(i = 0; i < len; i++){
	buf[i] = str[i];
    }

    for(i = 0; i < len; i++){
	buf[i] -= '0'; /* ASCII to number */
    }

    switch(len){
	case 4:
	    ret = buf[0]*1000 + buf[1]*100 + buf[2]*10 + buf[3];
	    break;
	case 5:
	    ret = buf[0]*10000 + buf[1]*1000 + buf[2]*100 + buf[3]*10 +buf[4];
	    break;
	case 6:
	    ret = buf[0]*100000 + buf[1]*10000 + buf[2]*1000 + buf[3]*100 + buf[4]*10 +buf[5];
	    break;
	default:
	    ret = 0;
	    break;
    }
    free(buf);

    return ret;
}

/* ***********************************************
 * Function	: main
 * argument	: argv[1] FILENAME
 * 		: argv[2] -b
 * 		: avgv[3] BUADRATE
 *
 * return	: FALSE or TRUE
 * ***********************************************/
int main(int argc, char **argv)
{
    int i;

    char *dev = "/dev/ttyS0"; /* 串口1 */
    unsigned long buad = 38400;/* 默认波特率 */


    int s_fd;
    int nread;
    unsigned char s_buf[132];

    char *filename;
    int f_fd;
    unsigned char f_buf[128];

    int trans_cnt;

    unsigned char pack_no, pack_no2, end_flag, tx_p;

    unsigned char xmodem_cmd;
    unsigned char err_cnt;
    unsigned char time_out_cnt;

    struct stat file_stat = {0};/* 文件属性 */

    printf("\n** AVR bootloader master for Linux");
    printf("\n** Version 0.01\n** Written by Mowin\n** All right reserved");
    printf("\n** Complied on %s at %s\n\n", __DATE__, __TIME__);


    /* 
     * argc    参数个数(连程序名本身)
     * argv[0] 程序名(第一个参数)
     * argv[1] 第一个参数
     * argv[2] 第二个参数
     * argv[3] 第三个参数
     * ...
     */

    /* 参数少于二个则未输入文件名 */
    if(argc < 2){
	fprintf(stderr, "avrboot: Usage: avrboot FILENAME [options]\n");
	fprintf(stderr, "avrboot: Options [-b] set the buadrate(default:38400). example:\n         specify buadrate=115200, usage: avrboot main.bin -b 115200\n");
	return FALSE;
    } 

    /* 只有文件名，波特率使用默认值38400 */
    else if(argc == 2){
	filename = argv[1]; 
    }

    /* 未输入波特率 */
    else if(argc == 3){
	fprintf(stderr, "avrboot: Usage: avrboot FILENAME [options]\n");
	fprintf(stderr, "avrboot: Options [-b] set the buadrate(default:38400). example:\n         specify buadrate=115200, usage: avrboot main.bin -b 115200\n");
	return FALSE;
    }

    /* 指定了波特率 */
    else if(argc == 4){
	filename = argv[1];

	if(strcmp(argv[2], "-b") == 0){
	    buad = str2hex(argv[3]);/* 将字符串转成数字 */
	} else {
	    fprintf(stderr, "avrboot: Option %s unknow\n", argv[2]);
	    return FALSE;
	}
    }


    /* 读写方式打开串口 */
    s_fd = open(dev, O_RDWR); 
    if(s_fd == FALSE){
	//fprintf(stderr, "avrboot: Can't open %s\n", dev);
	perror("avrboot: Can not open serial port");/* 用perror可以看到错误码 */
	return FALSE;
    }
    /* 设置串口速度 */
    if(set_speed(s_fd, buad) == FALSE){
	close(s_fd);
	return FALSE;
    }
    /* 设置停止位数据位及校验方式 */
    if(set_parity(s_fd, 8, 1, 'N') == FALSE){
	close(s_fd);
	return FALSE;
    } else {
	printf("avrboot: Opened port [%s] [%lu,8,1,N] fd=%d\n", dev, buad, s_fd);
    }

    /* 打开BIN文件 */
    f_fd = open(filename, O_RDONLY);/* 只读方式打开 */
    if(f_fd == FALSE){
	//fprintf(stderr, "avrboot: Can't open %s\n", filename);
	perror("avrboot: Can not open file");
	close(s_fd);
	return FALSE;
    } else {
	/* 获取文件属性 */
	if(fstat(f_fd, &file_stat) == FALSE){
	    perror("avrboot: Read file status error");
	    close(s_fd);
	    close(f_fd);
	    return FALSE;
	} else {
	    printf("avrboot: Opened file [%s], %ubytes fd=%d\n\n", filename, file_stat.st_size, f_fd);
	}
    }


    time_out_cnt = 0;
    fprintf(stderr, "avrboot: Transmit connected code | ");
    do{
	/* 发送联机密码 */
	xmodem_cmd = CONNECT_CODE;

	putc(xmodem_cmd, stderr);

	if(write(s_fd, &xmodem_cmd, 1) == FALSE){
	    fprintf(stderr, "avrboot: Port write error\n");
	    close(s_fd);
	    close(f_fd);
	    return FALSE;
	}

	/* 等待从机回应XMODEM_RWC time out = 1S*/
#if DEBUG
	printf("avrboot: Waiting for slave to respond\n");
#endif
	if(read(s_fd, &xmodem_cmd, 1) == FALSE){
	    //fprintf(stderr, "avrboot: Port write error\n");
	    perror("avrboot: Serial port read error");
	    close(s_fd);
	    close(f_fd);
	    return FALSE;
	}

	if(xmodem_cmd == XMODEM_RWC){
#if DEBUG
	    printf("avrboot: Receive XMODEM_RWC\n");
#endif
	    printf("\n");
	    break;
	}

	/* time out exit */
	if(++time_out_cnt >= 20){
	    printf("\navrboot: Slave has no response\n");
	    close(s_fd);
	    close(f_fd);
	    return FALSE;
	}

    } while(1);

    /* 发送开始控制字 XMODEM_SOH */
    xmodem_cmd = XMODEM_SOH;
    if(write(s_fd, &xmodem_cmd, 1) == FALSE){
	//fprintf(stderr, "avrboot: Port write error\n");
	perror("avrboot: Serial port write error");
	close(s_fd);
	close(f_fd);
	return FALSE;
    } else
	printf("avrboot: Starting data transmission\n\n");

    /* 开始发送数据包 */
    pack_no = 1;
    trans_cnt = 0;
    end_flag = 0;
    err_cnt = 0;
    fprintf(stderr, "Writing | ");
    for(;;){

	if(pack_no % 8 == 0){
	    putc('P', stderr);/* P = 1024B */
	}
	/* 从文件读取128B数据 */
	nread = read(f_fd, f_buf, 128);
	if(nread == FALSE){
	    //fprintf(stderr, "avrboot: %s read error\n", filename);
	    perror("avrboot: File read error");
	    close(s_fd);
	    close(f_fd);
	    return FALSE;
	} else {
	    /* 成功读取文件数据 */
	    crc16(f_buf, 128); /* 计算CRC校验 */

#if DEBUG
	    printf("avrboot: Pack[%d] nread %dB\n", pack_no, nread);
#endif

	    /* 计算已发送字节数 */
	    trans_cnt += nread;

	    /* 文件是否结束 */
	    if(nread < 128){
		printf("\n\navrboot: End of [%s], %dbytes transmited\n", filename, trans_cnt);
		end_flag = 1;/* 文件结束标志 */
	    }

#if 0
	    /* 清空发送缓存 */
	    for(i = 0; i < 132; i++){
		s_buf[i] = 0xff;
	    }
#endif

	    /* 将数据打包并填入发送缓冲区 */
	    /* 包序列号 */
	    pack_no2 = ~pack_no;
	    s_buf[0] = pack_no;
	    s_buf[1] = pack_no2;
	    /* 128B数据 */
	    for(tx_p = 2, i = 0; i < 128; i++){
		s_buf[tx_p++] = f_buf[i];
	    }
	    /* crch, crcl */
	    s_buf[tx_p++] = ch;
	    s_buf[tx_p] = cl;

	    /* 将数据写入串口送出 */
Re_trans:
	    if(write(s_fd, s_buf, sizeof(s_buf)) == FALSE){
		//fprintf(stderr, "avrboot: Port write error\n");
		perror("avrboot: Serial port write error");
		close(s_fd);
		close(f_fd);
		return FALSE;
	    } else {
#if DEBUG
		printf("avrboot: Pack[%d] transmited\n", pack_no);
#endif
	    }

	    /* 等待回应 */
	    do{
		if(read(s_fd, &xmodem_cmd, 1) == FALSE){
		    //fprintf(stderr, "avrboot: Port read error\n");
		    perror("avrboot: Serial port read error");
		    close(s_fd);
		    close(f_fd);
		    return FALSE;
		} else {
		    /* 收到回应 */
#if DEBUG
		    printf("%x ", xmodem_cmd);
#endif

		    if(xmodem_cmd == XMODEM_ACK){
#if DEBUG
			printf("avrboot: Receive XMODEM_ACK\n");
#endif
			/* 从机正确收到数据包 */
			if(end_flag == 0){
			    /* 文件未结束继续发送下一包数据 */
#if DEBUG
			    printf("avrboot: Next pack\n\n");
#endif
			    pack_no++;
			}
			err_cnt = 0;
			break;	
		    } 
		    else if(xmodem_cmd == XMODEM_NAK){
#if DEBUG
			printf("avrboot: Receive XMODEM_NAK\n");
#endif
			/* 从机要求重发数据包重发超过十次则放弃传输 */
			if(++err_cnt > 10){
			    printf("avrboot: Error too much, update failed\n");
			    end_flag = 1;
			    break;
			}

			/* 开始重发数据包 */
			xmodem_cmd = XMODEM_STX;
			if(write(s_fd, &xmodem_cmd, 1) == FALSE){
			    //fprintf(stderr, "avrboot: Port write error\n");
			    perror("avrboot: Serial port write error");
			    close(s_fd);
			    close(f_fd);
			    return FALSE;
			} else {
#if DEBUG
			    printf("avrboot: Transmit XMODEM_STX\n");
			    printf("avrboot: Re-transmit pack[%d]\n", pack_no);
#endif
			    goto Re_trans;
			}
		    }
		}

		/* 20秒等待超时退出升级 */
		if(++time_out_cnt >= 20){
		    printf("\navrboot: 20S time out, update failed\n");
		    close(s_fd);
		    close(f_fd);
		    return FALSE;
		}

	    }while(1);

	    if(end_flag == 1){
		/* 文件结束发送EOT */
		xmodem_cmd = XMODEM_EOT;
		if(write(s_fd, &xmodem_cmd, 1) == FALSE){
		    //fprintf(stderr, "avrboot: Port write error\n");
		    perror("avrboot: Serial port write error");
		    close(s_fd);
		    close(f_fd);
		    return FALSE;
		} else {
		    /* 结束传送 */
#if DEBUG
		    printf("avrboot: Transmit XMODEM_EOT\n");
#endif
		    break;
		}
	    } else {
		/* 开始下一包数据发送 */
		end_flag = 0;
		xmodem_cmd = XMODEM_STX;
		if(write(s_fd, &xmodem_cmd, 1) == FALSE){
		    //fprintf(stderr, "avrboot: Port write error\n");
		    perror("avrboot: Serial port write error");
		    close(s_fd);
		    close(f_fd);
		    return FALSE;
		} else {
#if DEBUG
		    printf("avrboot: Transmit XMODEM_STX\n");
#endif
		}
	    }
	}
    }


    if(err_cnt == 0)
	printf("avrboot: Update successed\n\n");
    else
	printf("avrboot: Update failed\n\n");

    close(s_fd);
    close(f_fd);

    return TRUE;
}


/* end of file */
